######################################################################################################
#
#   utility and plotting functions for the implementation of the oliker-prussner-method;
#   not all functions are actually used, some are only for debugging/testing;
#   for further details see the "main.py" file
#
#   by Yannick Höffner, Friedrich-Schiller-University Jena,
#   Bachelor Thesis WS 2024/25
#
#   based on the original paper from V.I.Oliker and L.D.Prussner:
#   "On the Numerical Solution of the Equation [...] and Its Discretizations, I"
#   Numerische Mathematik 54.3 (1989): 271-294
#
######################################################################################################
import numpy as np
import numpy.typing as npt

import matplotlib.pylab as plt
import matplotlib.tri as mtri


def show(triangles: npt.NDArray[np.int64], coordinates: npt.NDArray[np.float64],
         z_values: npt.NDArray[np.float64], figureNumber: np.int64 = 1, title: str = "") -> None:
    """plots the given function according to
    a triangulation and the coordinates of the
    triangulation vertices. Plots a triangle surface.

    Args:
        triangles (NDArray[int64]): list of triangles given as indices in
            the coords list
        coordinates (NDArray[float64]): x,y coordinates
        z_values (NDArray[float64]): z-value which should be plot to the vertices
        figureNumber (int64, optional): figure number where to plot the function.
            Defaults to 1.
        title (string, optional): title of the plot. Defaults empty string.
    """
    tri = mtri.Triangulation(
        coordinates[:, 0], coordinates[:, 1], triangles=triangles)
    fig = plt.figure(figureNumber)
    ax = fig.add_subplot(1, 1, 1, projection='3d')
    ax.set_title(title)
    ax.plot_trisurf(coordinates[:, 0], coordinates[:, 1],
                    z_values, triangles=tri.triangles, cmap=plt.cm.Spectral)
    plt.plot([0], [0], [0])


def plot_neigbourhood(index: np.int64, coordinates: npt.NDArray[np.float64],
                      z_values: npt.NDArray[np.float64], adjNext: npt.NDArray[np.int64],
                      fig_num: np.int64 = 1, max_alpha: np.float64 = None, alphas: npt.NDArray[np.float64] = None,
                      angles: npt.NDArray[np.float64] = None) -> None:
    """plots the given neighbourhood of a vertex with a potentially given alpha value

    Args:
        index (int64): the index of the vertex in the coords array
        coordinates (NDArray[float64]): x,y coordinates given [[x1,y1], ...]
        z_values (NDArray[float64]): values of the function at the vertices given as list with respect to the coords array
        adjNext (NDArray[int64]): adjacency information of the neighbourhood, see 'optri.getNeighborhoodCW' or 'optri.getNeighborhoodCCW'
            for further information
        fig_num (int64, optional): figure where to plot the neighbourhood. Defaults to 1.
        max_alpha (float64, optional): if you want to visualize the point where two neigbouring
            triangles fall into one plane. Defaults to None.
        alphas (NDArray[float64], optional): alpha values for each neighbouring vertex
        angles (NDArray[float64], optional): alpha values for each neighbouring vertex
    """
    fig = plt.figure(fig_num)
    ax = fig.add_subplot(1, 1, 1, projection='3d')
    ax.set_title(f"Neighbourhood of vertex {index}")

    ax.plot(coordinates[adjNext[1:-1], 0],
            coordinates[adjNext[1:-1], 1], z_values[adjNext[1:-1]], 'o')
    ax.plot(coordinates[index, 0],
            coordinates[index, 1], z_values[index], 'go')
    if max_alpha is not None and max_alpha != -np.inf:
        u_val_i = max_alpha
    else:
        u_val_i = z_values[index]
    ax.plot(coordinates[index, 0], coordinates[index, 1], u_val_i, 'ro')

    for i in range(1, adjNext.size-1):
        # plot line to neighbouring vertex
        ax.plot([coordinates[index, 0], coordinates[adjNext[i], 0]], [
                coordinates[index, 1], coordinates[adjNext[i], 1]], [
                    u_val_i, z_values[adjNext[i]]], marker=" ", color="black")
        # plot line between neighbouring vertices
        ax.plot([coordinates[adjNext[i], 0], coordinates[adjNext[i+1], 0]], [
                coordinates[adjNext[i], 1], coordinates[adjNext[i+1], 1]], [
                    z_values[adjNext[i]], z_values[adjNext[i+1]]], marker=" ", color="black")
        text: str = ""
        if alphas is not None:
            text = text + f"alp={alphas[i-1]}"
        if angles is not None:
            text = text + f" ang={angles[i-1]}"
        if len(text) > 0:
            text = f"{i} " + text
            ax.text(coordinates[adjNext[i], 0],
                    coordinates[adjNext[i], 1], z_values[adjNext[i]], text)



######################################################################################################
#
#   implementation of the oliker-prussner-method
#
#   by Yannick Höffner, Friedrich-Schiller-University Jena,
#   Bachelor Thesis WS 2024/25
#
#   based on the original paper from V.I.Oliker and L.D.Prussner:
#   "On the Numerical Solution of the Equation [...] and Its Discretizations, I"
#   Numerische Mathematik 54.3 (1989): 271-294
#
#
######################################################################################################
import numpy as np
import numpy.typing as npt
from numba import jit
from . import triangulation as optri
# import .oliker_prussner_coretriag as optri

########################################################################################################
#   IMPORTANT EQUATION CALCULATIONS
#   (see Appendix B in the thesis)
########################################################################################################


@jit(nopython=True)
def calculate_alpha(index: np.int64, coordinates: npt.NDArray[np.float64],
                    adjNext: npt.NDArray[np.int64], uValuesAdj: npt.NDArray[np.float64]) -> tuple[
                        npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """This function calculates the alpha values for the neighborhood of a vertex.
    See chapter 3.3.2 and appendix B.2 in the thesis for more information.

    Args:
        index (int64): the index under consideration
        coordinates (NDArray[float64]): array of vertex coordinates [[x1, y1], [x2, y2], ...]
        adjNext (NDArray[int64]): result of optri.getNeighborhoodCCW
        uValuesAdj (NDArray[float64]): current function values at the adjacent vertices

    Returns:
        alphas (NDArray[float64]): array of alpha values for the neighborhood
        angles (NDArray[float64]): array of outer angles for the neighborhood
    """
    alphas = np.zeros(adjNext.size-2, dtype=np.float64)
    angles = np.zeros(adjNext.size-2, dtype=np.float64)
    for i in range(1, adjNext.size-1):
        A_ij = np.append(coordinates[adjNext[i-1]], uValuesAdj[i-1])
        A_ij1 = np.append(coordinates[adjNext[i]], uValuesAdj[i])
        A_ij2 = np.append(coordinates[adjNext[i+1]], uValuesAdj[i+1])
        # set A_i[z] artificially to 0.0 to use cross products of p_i without z-influence
        A_i = np.append(coordinates[index], [0.0])

        Nbar_ij = np.cross(A_ij-A_i, A_ij1-A_i)
        Nbar_ij1 = np.cross(A_ij1-A_i, A_ij2-A_i)

        # calculating alpha for solution range
        # notation not from the paper
        D_alpha_j = np.array([A_ij1[1]-A_ij[1], A_ij[0]-A_ij1[0]])
        D_alpha_j1 = np.array([A_ij2[1]-A_ij1[1], A_ij1[0]-A_ij2[0]])

        a_alpha = Nbar_ij[:2]*Nbar_ij1[2] - Nbar_ij1[:2]*Nbar_ij[2]
        b_alpha = D_alpha_j1 * Nbar_ij[2] - D_alpha_j * Nbar_ij1[2]

        # sometimes one of the two components is zero in both a and b;
        # this depends on the position of the vertices
        if np.abs(b_alpha[1]) > 0.000001:
            alphas[i-1] = a_alpha[1]/b_alpha[1]
        elif np.abs(b_alpha[0]) > 0.000001:
            alphas[i-1] = a_alpha[0]/b_alpha[0]
        else:
            # this vertex lies on the line of the two adjacent vertices;
            # this vertex will never change the subgradient
            # since this vertex will always be on the face spanned by
            # Ai and the two adjacent vertices
            alphas[i-1] = np.inf

        # calculate outer angle of this point in the neigbourhood
        v_jj1 = A_ij - A_ij1
        v_jj2 = A_ij2 - A_ij1
        dot = v_jj1[0]*v_jj2[0] + v_jj1[1]*v_jj2[1]
        det = v_jj1[0]*v_jj2[1] - v_jj1[1]*v_jj2[0]
        angles[i-1] = np.degrees(np.arctan2(det, dot) % (2*np.pi))

    return alphas, angles


@jit(nopython=True)
def build_equation_parameters(index: np.int64, coordinates: npt.NDArray[np.float64],
                              adjNext: npt.NDArray[np.int64], uValuesAdj: npt.NDArray[np.float64]) -> tuple[
                                  np.float64, np.float64, np.float64]:
    """calculates the equation parameters a,b,c for the quadratic equation
    that represents the calculation of the Monge-Ampère-Measure for a vertex
    with unknown z-value.
    See chapter 3.3.2 and appendix B.1 in the thesis for more information.

    Args:
        index (int64): the index under consideration
        coordinates (NDArray[float64]): array of vertex coordinates [[x1, y1], [x2, y2], ...]
        adjNext (NDArray[int64]): result of optri.getNeighborhoodCCW
        uValuesAdj (NDArray[float64]): current function values at the adjacent vertices

    Returns:
        a,b,c (float64): the parameters a,b,c from the quadratic equation
    """
    a, b, c = 0, 0, 0
    for i in range(1, adjNext.size-1):
        # for detailed information about definition and naming of
        # variables in this loop, see the original Olliker Prussner Paper
        # on page 288 in the function new_z

        A_ij = np.append(coordinates[adjNext[i-1]], uValuesAdj[i-1])
        A_ij1 = np.append(coordinates[adjNext[i]], uValuesAdj[i])
        A_ij2 = np.append(coordinates[adjNext[i+1]], uValuesAdj[i+1])
        # set A_i[z] artificially to 0.0 to use cross products of p_i without z-influence
        A_i = np.append(coordinates[index], [0.0])

        Nbar_ij = np.cross(A_ij-A_i, A_ij1-A_i)
        Nbar_ij1 = np.cross(A_ij1-A_i, A_ij2-A_i)
        gamma_ij = Nbar_ij[2]
        gamma_ij1 = Nbar_ij1[2]

        N_ij = Nbar_ij/gamma_ij
        N_ij1 = Nbar_ij1/gamma_ij1

        D_ij = np.array([A_ij1[1]-A_ij[1], A_ij[0]-A_ij1[0]])/gamma_ij
        D_ij1 = np.array([A_ij2[1]-A_ij1[1], A_ij1[0]-A_ij2[0]])/gamma_ij1

        a_val = D_ij[0] * D_ij1[1] - D_ij[1] * D_ij1[0]
        b_val = D_ij[0] * N_ij1[1] - D_ij[1] * N_ij1[0]
        b_val = b_val + N_ij[0] * D_ij1[1] - N_ij[1] * D_ij1[0]
        c_val = N_ij[0] * N_ij1[1] - N_ij[1] * N_ij1[0]

        a = a + a_val
        b = b + b_val
        c = c + c_val

    a = -a/2
    b = -b/2
    c = -c/2
    return a, b, c


########################################################################################################
#   CORE FUNCTIONALITY OF THE OLIKER-PRUSSNER-METHOD
#   (see chapter 3 in the thesis)
########################################################################################################


def olliker_prussner_step(index: np.int64, coordinates: npt.NDArray[np.float64],
                          triangles: npt.NDArray[np.int64], adjacancies: npt.NDArray,
                          count_adj: npt.NDArray[np.int64], old_z_values: npt.NDArray[np.float64],
                          measure: npt.NDArray[np.float64]) -> np.float64:
    """This function implements one z-update step of the Oliker-Prussner-Method for a single vertex.
    This function basically implements algorithm 3 of the thesis.
    See chapter 3.3.2 in the thesis for more information.

    Args:
        index (int64): the index under consideration
        coordinates (NDArray[float64]): array of vertex coordinates [[x1, y1], [x2, y2], ...]
        triangles (NDArray[int64]): see optri
        adjacancies (NDArray[any]): see optri
        count_adj (NDArray[int64]): see optri
        old_z_values (NDArray[float64]): array of current function values at the vertices
        measure (NDArray[float64]): array of the pointwise measure mu, that we want to achieve (for every vertex)

    Returns:
        z (float64): updated z-value for the vertex
    """
    current_value = old_z_values[index]
    adjNext, _ = optri.get_neighborhood_CCW(
        index, triangles, adjacancies,  count_adj)
    while True:
        uValuesAdj = old_z_values[adjNext]
        ###########################################################
        # calculate solution range (alpha)
        ###########################################################
        alphas, angles = calculate_alpha(
            index, coordinates, adjNext, uValuesAdj)
        max_alpha_idx = -1
        max_alpha = -np.inf
        for i in range(alphas.size):
            # only if the outer angle is less than 180°,
            # then this vertex will not be on the convex hull
            # if you go down deeper in the z direction
            if angles[i] < 180:
                if alphas[i] > max_alpha and alphas[i] < current_value + 0.000001:
                    max_alpha = alphas[i]
                    max_alpha_idx = i

        ###########################################################
        # build sub gradient equation
        ###########################################################
        # coefficients for the quadratic equation az^2+bz+c
        a, b, c = build_equation_parameters(
            index, coordinates, adjNext, uValuesAdj)
        ###########################################################
        # solve sub gradient equation
        ###########################################################
        # equation -mu_i = a * z^2 + b * z + c

        c = c + measure[index]

        # equation  0 = a * z^2 + b * z + c
        # details about this numerically stable quadratic solver:
        # https://math.stackexchange.com/q/2007723
        u = -b - np.copysign(np.sqrt(b*b-4*a*c), b)
        sol = np.array([u / (2*a), (2*c) / u])

        sol2 = sol[(sol < current_value) & (
            sol > max_alpha) & np.isfinite(sol)]

        if len(sol2) > 0:
            return np.max(sol2)
        if max_alpha == -np.inf:
            # i currently don't know why this case happens from time to time,
            # but we just skip the vertex if it happens and let the "self healing"
            # property of global operation of the Oliker-Prussner-Method do its job
            print("This should not happen, eq: a,b,c: ", a, b, c)
            return current_value
        # remove the vertex, that will not be on the convex hull
        # and repeat the calculation
        if max_alpha_idx == 0:
            adjNext = np.delete(adjNext, adjNext.size-1)
            adjNext = np.delete(adjNext, 1)
            adjNext = np.append(adjNext, adjNext[1])
        elif max_alpha_idx == alphas.size-1:
            adjNext = np.delete(adjNext, adjNext.size-2)
            adjNext = np.delete(adjNext, 0)
            adjNext = np.append(adjNext, adjNext[1])
        else:
            adjNext = np.delete(adjNext, max_alpha_idx+1)


def oliker_prussner_sceme(coordinates: npt.NDArray[np.float64], boundary_mask: npt.NDArray[np.bool_],
                          initial_triags: npt.NDArray[np.int64], initial_values: npt.NDArray[np.float64],
                          measure: npt.NDArray[np.float64], infostring: str, max_iterations: np.int64) -> tuple[
                              npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.int64], np.int64]:
    """This function implements the Oliker-Prussner-Method.
    See chapter 3 in the thesis or the original paper from Oliker and Prussner for more information.

    Args:
        coordinates (NDArray[float64]): array of vertex coordinates [[x1, y1], [x2, y2], ...]
        boundary_mask (NDArray[bool]): boundary mask of the coordinate list
        initial_triags (NDArray[int64]): triangle array of start function; see optri
        initial_values (NDArray[float64]): function values of the start function at the vertices
        measure (NDArray[float64]): the Monge-Ampère-Measure that we want to achieve
        infostring (string): string that will be displayed in the console
        max_iterations (int64): maximum number of iterations

    Returns:
        calculated_z_values (NDArray[float64]): updated function values
        calculated_triags (NDArray[int64]): updated triangulation array
        nbr_iterations (int64): number of iterations needed
    """
    np.set_printoptions(precision=16)
    monge_ampere_precision = 1e-6
    nbr_iterations = 0

    # calculate adjacencie information from the triangulation
    count_adj, adjacencies, _ = optri.find_adjacencies(
        coordinates, initial_triags)
    # rename the variables for better readability
    calculated_z_values = initial_values
    calculated_triags = initial_triags

    for repetitions in range(max_iterations):
        innerIndices = np.arange(coordinates.shape[0])[~boundary_mask]
        new_z = np.copy(calculated_z_values)

        ###########################################################
        # adjust z values by Monge-Ampère-Measure calculation
        ###########################################################

        calculatedIndices = 0
        for index in innerIndices:
            monge_ampere_measure = optri.calculate_monge_ampere_measure(
                index, coordinates, calculated_triags, adjacencies, count_adj, calculated_z_values)
            if np.abs(monge_ampere_measure-measure[index]) < monge_ampere_precision:
                # the curvature/bulge of the function is already correct at this vertex
                continue
            new_z[index] = olliker_prussner_step(index, coordinates, calculated_triags, adjacencies,
                                                 count_adj, calculated_z_values, measure)
            calculatedIndices = calculatedIndices + 1
        if calculatedIndices == 0:
            # break iteration if Monge-Ampère-Measure is equal to the given measure mu (up to tollerance)
            # for all inner vertices
            print("no more update done, iteration finished")
            nbr_iterations = repetitions
            break

        ###########################################################
        # globally overrite the z values
        ###########################################################
        calculated_z_values = new_z

        ###########################################################
        # recalculate the triangulation with the new z values
        ###########################################################

        hullP = np.vstack((coordinates.T, [calculated_z_values])).T
        hull = optri.get_lower_convex_hull(hullP)
        print("lower convex hull calculated")

        remainingIndices = np.setdiff1d(innerIndices, hull.vertices)
        print("remaining points: ", remainingIndices)

        points, triangles = optri.refine_triangulation(
            hull.simplices, hull.points, remainingIndices)
        calculated_z_values = np.copy(points[:, 2])

        calculated_triags = optri.sort_triangles_CCW(
            coordinates, np.copy(triangles))
        count_adj, adjacencies, _ = optri.find_adjacencies(
            coordinates, calculated_triags)
        ###########################################################
        # print iteration information
        ###########################################################
        print(f"{infostring} it:{repetitions}")

    return calculated_z_values, calculated_triags, nbr_iterations



#####################################################################################################
#
#   initialisation of the oliker-prussner-method
#
#   by Yannick Höffner, Friedrich-Schiller-University Jena,
#   Bachelor Thesis WS 2024/25
#
#   based on the original paper from V.I.Oliker and L.D.Prussner:
#   "On the Numerical Solution of the Equation [...] and Its Discretizations, I"
#   Numerische Mathematik 54.3 (1989): 271-294
#
######################################################################################################
import numpy as np
import numpy.typing as npt
from typing import Callable
from scipy.spatial import Delaunay
from . import triangulation as optri


########################################################################################################
#   MEASURE MU INITIALIZATION
#   (see chapter 3.2.3 in the thesis)
########################################################################################################


def compute_MAM_from_grid(adjacancies: npt.NDArray, triangles: npt.NDArray[np.int64],
                          countADJ: npt.NDArray[np.int64], coordinates: npt.NDArray[np.float64],
                          boundaryMask: npt.NDArray[np.bool_], rhs_f: Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]) -> npt.NDArray[np.float64]:
    """calculates a possible Monge-Ampère-Measure from an initial grid by setting the
    local measure of a vertex to the integral over the classical FEM hat functions
    on adjoint triangles

    Args:
        adjacancies (NDArray[Any]): adjacencies list of initial triangulation;
            see optri.findAdjacencies for further information
        triangles (NDArray[int64]): triangles list of initial triangulation;
            see optri.findAdjacencies for further information
        countADJ (NDArray[int64]): adjacencies count of initial triangulation;
            see optri.findAdjacencies for further information
        coordinates (NDArray[float64]): array [[x1,y1],[x2,y2],...] with coordinate data of
            the vertices
        boundaryMask (NDArray[float64]): mask for coordinates with 1 if the desired vertex
            is an boundary point and 0 if it is an interior point
        rhs_f (callable): function for the right hand side of the Monge-Ampère-Equation

    Returns:
        mu_from_grid (NDArray[float64]): measure Monge-Ampère-Measure computed at every vertex
    """
    mu_from_grid = np.zeros(coordinates.shape[0], dtype=np.float64)
    for i in range(coordinates.shape[0]):
        if boundaryMask[i]:
            continue  # skip boundary points
        neighbourhood = adjacancies[i]
        next = np.array([1, 2, 0])
        for j in range(countADJ[i]):
            triag = triangles[neighbourhood[j][0]]
            # vertex0 is equivalent to i
            vertex1 = triag[next[neighbourhood[j][1]]]
            vertex2 = triag[next[next[neighbourhood[j][1]]]]

            # implementing second order quadrature from Saleri p. 415
            midpoint1 = np.array([(coordinates[i]+coordinates[vertex1])/2])
            midpoint2 = np.array([(coordinates[i]+coordinates[vertex2])/2])
            vec1 = coordinates[vertex1]-coordinates[i]
            vec2 = coordinates[vertex2]-coordinates[i]
            area = np.abs(vec1[0] * vec2[1] - vec1[1] * vec2[0])/2.0
            mu_from_grid[i] = mu_from_grid[i] + \
                (rhs_f(midpoint1)[0]+rhs_f(midpoint2)[0])*area/6.0
    return mu_from_grid


def compute_MA_measure(coordinates: npt.NDArray[np.float64], boundaryMask: npt.NDArray[np.bool_],
                       rhs_f: Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]]) -> npt.NDArray[np.float64]:
    """calculates a possible Monge-Ampère-Measure from an initial vertices by applying
    a delaunay triangulation and using "compute_mu_from_grid"
    See chapter 3.2.3 in the thesis for further information.

    Args:
        coordinates (NDArray[float64]): array [[x1,y1],[x2,y2],...] with coordinate data of
            the vertices
        boundaryMask (NDArray[bool]): mask for coordinates with 1 if the desired vertex
            is an boundary point and 0 if it is an interior point
        rhs_f (callable): function for the right hand side of the Monge-Ampère-Equation

    Returns:
        mu (NDArray[float64]): Monge-Ampère-Measure computed at every vertex
    """
    tri = Delaunay(coordinates).simplices
    count, adj, _ = optri.find_adjacencies(coordinates, tri)

    return compute_MAM_from_grid(adj, tri, count, coordinates, boundaryMask, rhs_f)


########################################################################################################
#   TRIANGULATION INITIALIZATION
#   (see chapter 3.2.2 in the thesis)
########################################################################################################


def planarity_check(points: npt.NDArray[np.float64]) -> np.bool_:
    """checks whether a given point cloud is planar

    Args:
        points (NDArray[float64]): 3d point cloud [[x1,y1,z1],[x2,y2,z2],...]

    Returns:
        bool: true if the point cloud is planar, false otherwise
    """
    p0 = points[0]
    p1 = points[1]
    p2 = points[-1]
    v1 = p1-p0
    v2 = p2-p0
    n = np.cross(v1, v2)
    n = n/np.linalg.norm(n)
    remaining_points = points[2:-1]
    remaining_points = remaining_points-p0
    dotp_vals = np.dot(remaining_points, n)
    return np.all(np.isclose(dotp_vals, 0.0))


def create_init_triangulation(coords: npt.NDArray[np.float64], boundary_mask: npt.NDArray[np.bool_], boundary_values: npt.NDArray[np.float64]) -> tuple[npt.NDArray[np.int64], npt.NDArray[np.float64]]:
    """Creates the initial function graph in the space W_M for the Oliker-Prussner-Method.
    This function is the lower boundary of the convex hull of the boundary points (the function
    values of the boundary points are given through the boundary condition of the Monge-
    Ampère-Equation). See also chapter 3.2.2 in the thesis for further information.

    THE COORDINATES/BOUNDARY MASK MUST BE SORTED WITH BOUNDARY POINTS FIRST!

    Args:
        coords (NDArray[float64]): list of x,y coordinates of the grid points, 
            where the boundary points should be sorted to the beginning of the list
        boundary_mask (NDArray[bool]): mask for coords; valued 1 if the desired vertex is a boundary point, otherwise 0;
            boundary points should be sorted to the beginning of the list
        boundary_values (NDArray[float64]): function values at the boundary points (all interior points should be 0!)
            boundary_values should not contain nan values.

    Returns:
        triags: the initial triangulation of the domain

        init_u_values: the initial function values at the grid points
    """
    # combine the boundary points with their function values given by the boundary condition
    hull_points = np.vstack(
        (coords[boundary_mask].T, [boundary_values[boundary_mask]])).T

    indices = np.arange(coords.shape[0])
    if planarity_check(hull_points):
        # 3d version of scipy convex hull (used in optri.getLowerConvexHull)
        # is not working with planar 3d point cloud
        hull_points_2d = coords[boundary_mask]
        delauney2d = Delaunay(hull_points_2d)
        # set the z coordinate of the unused points to 0 because
        # the z coordinate will be calculated in optri.refineTriangulation
        remainingIndices = np.setdiff1d(indices, delauney2d.simplices)
        init_u_values = np.vstack(
            (coords.T, np.repeat(0, coords.shape[0]))).T
        init_u_values[delauney2d.simplices,
                      2] = boundary_values[delauney2d.simplices]
        # complete the convex hull of some/all boundary points to
        # a triangulation of the whole domain (every vertex is a grid point)
        init_u_values, triags = optri.refine_triangulation(
            delauney2d.simplices, init_u_values, remainingIndices)
    else:
        lowerHull = optri.get_lower_convex_hull(hull_points)
        # set the z coordinate of the unused points to 0 because
        # the z coordinate will be calculated in optri.refineTriangulation
        remainingIndices = np.setdiff1d(indices, lowerHull.vertices)
        init_u_values = np.vstack(
            (coords.T, np.repeat(0, coords.shape[0]))).T
        init_u_values[lowerHull.vertices,
                      2] = boundary_values[lowerHull.vertices]
        # complete the convex hull of some/all boundary points to
        # a triangulation of the whole domain (every vertex is a grid point)
        init_u_values, triags = optri.refine_triangulation(
            lowerHull.simplices, init_u_values, remainingIndices)

    triags = optri.sort_triangles_CCW(
        np.copy(coords), np.copy(triags))

    return triags, init_u_values[:, 2]



#####################################################################################################
#
#   functions related to the management of the triangulation for the oliker-prussner-method,
#   not all functions are actually used, some are for testing/debugging
#   for further details see the "main.py" file
#
#   by Yannick Höffner, Friedrich-Schiller-University Jena,
#   Bachelor Thesis WS 2024/25
#
#   based on the original paper from V.I.Oliker and L.D.Prussner:
#   "On the Numerical Solution of the Equation [...] and Its Discretizations, I"
#   Numerische Mathematik 54.3 (1989): 271-294
#
######################################################################################################
import numpy as np
import numpy.typing as npt
from numba import jit
from scipy.spatial import ConvexHull, Delaunay


@jit(nopython=True)
def intersect_line_triangle(point: npt.NDArray[np.float64], p1: npt.NDArray[np.float64],
                            p2: npt.NDArray[np.float64], p3: npt.NDArray[np.float64]) -> tuple[
        npt.NDArray[np.float64] | None, npt.NDArray[np.float64] | None, npt.NDArray[np.float64] | None]:
    """calculates the intersection of a vertical line
    (directional vector [0,0,1]) starting at 'point' with
    a triangle in R^3 build by the vertices 'p1','p2','p3'

    Args:
        point (NDArray[float64]): x, y coordinates of the starting point of the
            ray, z is set to 0
        p1 (NDArray[float64]): x,y,z coordinates of first triangle vertex
        p2 (NDArray[float64]): x,y,z coordinates of second triangle vertex
        p3 (NDArray[float64]): x,y,z coordinates of third triangle vertex

    Returns:
        N (NDArray[float64]): the normal vector of the triangle (R^3), None if N[2] = 0
        p (NDArray[float64]): the intersection point of the line and triangle, None if
            intersection does not exist or N[2] = 0
        dist (NDArray[float64]): the signed distances of the plane intersection point to the
            different triangle sides, None if N[2] = 0
    """
    # TODO: maybe adjust the algorithm
    # we are basically performing lots of ray/triangle intersections
    # wich can be done better with the  Möller and Trumbore algorithm
    #
    # there also exist (seemingly) even better algorithms:
    # Doug Baldwin and Michael Weber, Fast Ray-Triangle Intersections by Coordinate Transformation
    # https://jcgt.org/published/0005/03/03/

    # current code from stackoverflow
    # https://stackoverflow.com/questions/54143142/3d-intersection-between-segment-and-triangle
    # by user: Mauricio Cele Lopez Belon

    # triangle normal vector
    N = np.cross((p2 - p1), (p3 - p1))

    if abs(N[2]) < 0.000001:
        return None, None, None
    # intersection point of triangle plane and line
    p = point - (-np.dot(N, p1) + np.dot(N, point)) / \
        N[2] * np.array([0, 0, 1])
    # explanaition: N[2] == np.dot(N, [0,0,1])

    # normal vectors of triangle sides within the triangle plane
    N12 = np.cross((p2 - p1), N)
    N23 = np.cross((p3 - p2), N)
    N31 = np.cross((p1 - p3), N)

    # distance of intersection point to triangle sides
    Dist1 = np.dot((p - p1), N12) / np.linalg.norm(N12)
    Dist2 = np.dot((p - p2), N23) / np.linalg.norm(N23)
    Dist3 = np.dot((p - p3), N31) / np.linalg.norm(N31)
    if Dist1 < 0.00001 and Dist2 < 0.00001 and Dist3 < 0.00001:
        return N, p, np.array((Dist1, Dist2, Dist3))
    return N, None, np.array((Dist1, Dist2, Dist3))


def sort_triangles_CCW(coordinates: npt.NDArray[np.float64], triags: npt.NDArray[np.int64]) -> npt.NDArray[np.int64]:
    """sorts the vertices of the given triangles, such that
    they are in ccw order

    Args:
        coordinates (NDArray[float64]): array of vertex coordinates [[x1, y1], [x2, y2], ...]
        triags (NDArray[int64]): array of triangle vertex indices [ [t1v1, t1v2, t1v3], [t2v1, t2v2, t2v3], ...]

    Returns:
        triags (NDArray[int64]): the sorted triangle array
    """
    for i in range(triags.shape[0]):
        triangle = triags[i]
        coords = coordinates[triangle]
        center = np.mean(coords, axis=0)
        localCoords = coords - center
        arg = np.arctan2(localCoords[:, 1], localCoords[:, 0])
        idx_CCW = np.argsort(arg)
        triags[i] = triags[i][idx_CCW]
    return triags


def find_adjacencies(coordinates: npt.NDArray[np.float64], triags: npt.NDArray[np.int64]) -> tuple[
        npt.NDArray[np.int64], npt.NDArray, npt.NDArray[np.int64]]:
    """finds the adjacent triangle of the vertices and stores
    the indices in a linked list

    the adjacent triangles inside the linked list are stored in
    cw order and ccw order.
    FOR PROPPER SORTING, TRIANGLE VERTICES MUST BE IN CCW ORDER

    Args:
        coordinates (NDArray[float64]): array of vertex coordinates [[x1, y1], [x2, y2], ...]
        triags (NDArray[int64]): array of triangle vertex indices [ [t1v1, t1v2, t1v3], [t2v1, t2v2, t2v3], ...]

    Returns:
        count (NDArray[int64]): array where count[i] is the number of adjacent triangles to vertex i
        adj (NDArray[any]): array where adj[i] is the linked list of adjacent triangles to vertex i.
            The k-th entry in this linked list is:\n
            adj[i][k][0] is the triangle index;\n
            adj[i][k][1] is the index of vertex i in triangle adj[i][0];\n
            adj[i][k][2] is the index in adj[i] where the cw next triangle appears;\n
            adj[i][k][3] is the index in adj[i] where the ccw next triangle appears;
        elementIdx (NDArray[int64]): array where elementIdx[i] contains the indices of the adjacencie-list
            of the vertices of the i-th triangle
    """
    maxNumberOfAdjacencies = 500
    # TODO: adjust the maximum number of vertices when domain changes
    adj = np.zeros(
        (coordinates.shape[0], maxNumberOfAdjacencies, 4), dtype=np.int32)
    # stores the number of adj vertices
    count = np.zeros((coordinates.shape[0]), dtype=np.int32)
    # stores the location of the triangles in the adj-array
    elementIdx = np.zeros(triags.shape)

    for i in range(triags.shape[0]):
        for j in range(3):
            k = triags[i][j]
            # for the k-th vertex the count[k]-th adjacent triangle is the i-th triangle
            adj[k][count[k]][0] = i
            # for the k-th vertex and the count[k]-th adjacent triangle,
            # the vertex k is stored as j-th edge of the triangle
            adj[k][count[k]][1] = j
            # store element idx
            elementIdx[i, j] = count[k]
            # increment the number of adjacend triangles
            count[k] = count[k]+1
    # currently the adjacent triangles are sorted by there appearence in elements.
    # now we need to sort the adjacent triangles, such that we can iterate through them
    # in clockwise order.
    next = np.array([1, 2, 0])
    for i in range(coordinates.shape[0]):
        # i = 7
        cnt = np.arange(count[i])
        # get list of all adjacent vertices
        adjInd = triags[adj[i, cnt, 0], next[adj[i, cnt, 1]]]
        # get relative angular coordinates
        localCoords = coordinates[adjInd] - coordinates[i]
        arg = np.arctan2(localCoords[:, 1], localCoords[:, 0])
        # sort angle descending
        idx_CW = np.flip(np.argsort(arg))
        idx_CCW = np.argsort(arg)
        idx_roll_CW = np.roll(idx_CW, +1)
        idx_roll_CCW = np.roll(idx_CCW, +1)
        temp_CW = np.zeros((count[i]))
        temp_CCW = np.zeros((count[i]))
        temp_CW[idx_roll_CW] = idx_CW
        temp_CCW[idx_roll_CCW] = idx_CCW
        # store index to next CW triangle
        adj[i, cnt, 2] = temp_CW
        # store index to next CCW triangle
        adj[i, cnt, 3] = temp_CCW
        # break
    return count, adj, elementIdx


def get_neighborhood_CW(index: np.int64, triangles: npt.NDArray[np.int64],
                        adjacencies: npt.NDArray, countADJ: npt.NDArray[np.int64]) -> tuple[
        npt.NDArray[np.int64], npt.NDArray[np.int64]]:
    """reads out the adjacencies of a triangle mesh and returns a copy
    of the neighboring triangles and vertices in cw order, of the vertex 'index'

    Args:
        index (int64): the vertices-index where the neighborhood should be calculated
        triangles (NDArray[int64]): triangle description of function graph
        adjacancies (NDArray[any]): adjacencies relationship of the triangulation
        countADJ (NDArray[int64]): number of adjacent triangles for each vertex

    Returns:
        adjNext (NDArray[int64]): list of adjacent vertices in cw order
        triags (NDArray[int64]): list of adjacent triangles in cw order
    """
    CWsorting = adjacencies[index, :countADJ[index], 2]
    temp = np.zeros(countADJ[index], dtype=np.int32)
    for i in range(1, countADJ[index]):
        temp[i] = CWsorting[temp[i-1]]
    CWsorting = temp
    adj = adjacencies[index, :countADJ[index], :]
    triags = triangles[adj[:, 0]]
    next = np.array([1, 2, 0])
    adjNext = triags[np.arange(countADJ[index]), next[adj[:, 1]]]

    triags = triags[CWsorting]
    adjNext = adjNext[CWsorting]

    adjNext = np.append(adjNext, adjNext[0])
    adjNext = np.append(adjNext, adjNext[1])

    return np.copy(adjNext), np.copy(triags)


def get_neighborhood_CCW(index: np.int64, triangles: npt.NDArray[np.int64],
                         adjacencies: npt.NDArray, countADJ: npt.NDArray[np.int64]) -> tuple[
                             npt.NDArray[np.int64], npt.NDArray[np.int64]]:
    """reads out the adjacencies of a triangle mesh and returns a copy
    of the neighboring triangles and vertices in ccw order, of the vertex 'index'

    Args:
        index (int64): the vertices-index where the neighborhood should be calculated
        triangles (NDArray[int64]): triangle description of function graph
        adjacancies (NDArray[any]): adjacencies relationship of the triangulation
        countADJ (NDArray[int64]): number of adjacent triangles for each vertex

    Returns:
        adjNext (NDArray[int64]): list of adjacent vertices in ccw order
        triags (NDArray[int64]): list of adjacent triangles in ccw order
    """
    CCWsorting = adjacencies[index, :countADJ[index], 3]
    temp = np.zeros(countADJ[index], dtype=np.int32)
    for i in range(1, countADJ[index]):
        temp[i] = CCWsorting[temp[i-1]]
    CCWsorting = temp
    adj = adjacencies[index, :countADJ[index], :]
    triags = triangles[adj[:, 0]]
    next = np.array([1, 2, 0])
    adjNext = triags[np.arange(countADJ[index]), next[adj[:, 1]]]

    triags = triags[CCWsorting]
    adjNext = adjNext[CCWsorting]

    adjNext = np.append(adjNext, adjNext[0])
    adjNext = np.append(adjNext, adjNext[1])

    return np.copy(adjNext), np.copy(triags)


def calculate_monge_ampere_measure(index: np.int64, coordinates: npt.NDArray[np.float64],
                                   triangles: npt.NDArray[np.int64], adjacencies: npt.NDArray,
                                   count_adj: npt.NDArray[np.int64],  z_values: npt.NDArray[np.float64]) -> np.float64:
    """calculated the Monge-Ampère-Measure of the graph at vertex "index".
    see 'oliker_prussner_init' for more details on the input values

    Args:
        index (int64): the vertices-index where the Monge-Ampère-Measure should be calculated
        coordinates (NDArray[float64]): x and y coordinates of the vetices
        triangles (NDArray[int64]): triangle description of function graph
        adjacencies (NDArray[any]): adjacencies relationship of the triangulation
        count_adj (NDArray[int64]): number of adjacent triangles for each vertex
        z_values (NDArray[float64]): z coordinates of vertices

    Returns:
        monge_ampere_measure (float64): Monge-Ampère-Measure of vertex 'index'
    """
    adjNext, _ = get_neighborhood_CCW(
        index, triangles, adjacencies, count_adj)

    uValuesAdj = z_values[adjNext]
    coordsAdj = coordinates[adjNext]

    monge_ampere_measure = 0
    for k in range(1, adjNext.size-1):

        p_ij = np.append(coordsAdj[k-1], uValuesAdj[k-1])
        p_ij1 = np.append(coordsAdj[k], uValuesAdj[k])
        p_ij2 = np.append(coordsAdj[k+1], uValuesAdj[k+1])
        p_i = np.append(coordinates[index], z_values[index])

        Nbar_ij = np.cross(p_ij-p_i, p_ij1-p_i)
        Nbar_ij1 = np.cross(p_ij1-p_i, p_ij2-p_i)
        N_ij = np.array([Nbar_ij[0], Nbar_ij[1]])/Nbar_ij[2]
        N_ij1 = np.array([Nbar_ij1[0], Nbar_ij1[1]])/Nbar_ij1[2]

        monge_ampere_measure = monge_ampere_measure + \
            N_ij[0] * N_ij1[1] - N_ij[1] * N_ij1[0]

    monge_ampere_measure = monge_ampere_measure/2
    return np.abs(monge_ampere_measure)


@jit(forceobj=True)
def get_lower_convex_hull(hullPoints: npt.NDArray[np.float64]) -> ConvexHull:
    """calculated the convex hull of a given point cloud in R^3 and
    removes all triangles that do not belong to the lower part of the
    convex hull

    Args:
        hullPoints (NDArray[float64]): x, y, z coordinates of the points from which the hull
            should be build

    Returns:
        hull (scipy.spatial.ConvexHull): the convex hull object with modified simplices
    """
    hull = ConvexHull(hullPoints)  # , qhull_options="Q0")
    trianglePoints = hullPoints[hull.simplices]

    point = np.array([0.0, 0.0, 0.0])
    # remove vertical triangles
    for j in range(trianglePoints.shape[0]-1, -1, -1):
        triangle = trianglePoints[j]
        normalVec, intersect, _ = intersect_line_triangle(
            point, triangle[0], triangle[1], triangle[2])
        if normalVec is None:
            trianglePoints = np.delete(
                trianglePoints, j, 0)
            hull.simplices = np.delete(
                hull.simplices, j, 0)

    # remove upper triangles
    i = trianglePoints.shape[0]-1
    while i > -1:
        triangleI = trianglePoints[i]
        point = np.mean(triangleI, axis=0)
        _, intersectI, _ = intersect_line_triangle(
            point, triangleI[0], triangleI[1], triangleI[2])
        for j in range(trianglePoints.shape[0]-1, -1, -1):
            if i == j:
                continue
            triangleJ = trianglePoints[j]
            _, intersectJ, _ = intersect_line_triangle(
                point, triangleJ[0], triangleJ[1], triangleJ[2])
            if intersectJ is not None:
                if intersectJ[2] > intersectI[2] and np.abs(intersectJ[2]-intersectI[2]) > 1e-6:
                    trianglePoints = np.delete(
                        trianglePoints, j, 0)
                    hull.simplices = np.delete(
                        hull.simplices, j, 0)
                    if j < i:
                        i = i - 1
        i = i-1
    return hull


def find_points_on_lower_convex_hull(hullPoints: npt.NDArray[np.float64],
                                     projPoints: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """generates a convex Hull from the points 'hullPoints' and
    calculates the projection of the points 'projPoints' onto the
    lower boundary of the convex hull (in z-direction)

    technically perfoming a ray tracing from a point in 'projPoints'
    by direction vector [0,0,1] and calculating intersections with the
    triangles that form the convex hull.

    DOES NOT CHECK, WHETHER AN INTERSECTION WITH THE CONVEX HULL EXISTS!

    Args:
        hullPoints (NDArray[float64]): array [[x1, y1, z1], ...] of 3D points to form the
            convex hull from
        projPoints (NDArray[float64]): array [[x1, y1], ...] of 2D points to be projected
            onto the convex hull

    Returns:
        NDArray[float64]: array [z1, z2, ...] of z-values to the corresponding 2D input points
    """
    hull = ConvexHull(hullPoints)
    trianglePoints = hullPoints[hull.simplices]

    finalZValues = np.zeros((projPoints.shape[0]))
    for i in range(projPoints.shape[0]):
        point = projPoints[i]
        finalPoint = np.array([point[0], point[1], 0.1e100])

        for j in range(trianglePoints.shape[0]-1, -1, -1):
            triangle = trianglePoints[j]
            _, intersect, _ = intersect_line_triangle(
                np.append(point, [0]), triangle[0], triangle[1], triangle[2])
            if intersect is not None:
                if intersect[2] < finalPoint[2]:
                    finalPoint = intersect
        finalZValues[i] = finalPoint[2]

    return finalZValues


@jit(forceobj=True)
def refine_triangulation(hullTriangles: npt.NDArray[np.int64], hullPoints: npt.NDArray[np.float64],
                         remainingIndices: npt.NDArray[np.int64]) -> tuple[npt.NDArray[np.float64],
                                                                           npt.NDArray[np.int64]]:
    """adds the remaining points to the triangulation of the convex hull
    by calculating the intersection of the points with the triangles of the
    convex hull. The new points are added to the hullPoints array and the
    new triangles are added to the hullTriangles array.

    Args:
        hullTriangles (NDArray[int64]): list of all triangles of the lower convex hull 
        hullPoints (NDArray[float64]): list of all 3d-points used to calculate the convex hull
            (must also contain the points that are currently not vertices of the convex hull!)
        remainingIndices (NDArray[int64]): list of indices of the points that are not part of the convex hull

    Returns:
        points (NDArray[float64]): points array with updated z-data
        triags (NDArray[int64]): triangle array with updated triangles
    """
    points = np.copy(hullPoints)
    triags = np.array([[0, 0, 0]])
    for i in range(hullTriangles.shape[0]):
        pi1, pi2, pi3 = hullTriangles[i]
        p1, p2, p3 = hullPoints[[pi1, pi2, pi3]]
        found_points = [p1, p2, p3]
        found_indices = [pi1, pi2, pi3]
        for j in remainingIndices:
            point = points[j]
            _, p, _ = intersect_line_triangle(
                np.array([point[0], point[1], 0]), p1, p2, p3)
            if p is not None:
                points[j, 2] = p[2]
                found_points.append(np.array([point[0], point[1], p[2]]))
                found_indices.append(j)
        found_points = np.array(found_points)
        found_indices = np.array(found_indices)

        tri = Delaunay(found_points[:, :2])
        # indices of tri are with respect to the enumeration of found_points
        # but we need the indices with respect to the enumeration of hullPoints
        tri_correctec = found_indices[tri.simplices]
        # scan triangulation for degenerated triangles
        for k in range(tri_correctec.shape[0]-1, -1, -1):
            local_coords = points[tri_correctec[k]].copy()
            local_coords[:, 2] = 1
            if np.isclose(np.linalg.det(local_coords), 0.0):
                tri_correctec = np.delete(tri_correctec, k, 0)
        triags = np.append(triags, tri_correctec, axis=0)

    triags = np.delete(triags, 0, axis=0)
    return points, triags
